/**********************************************************************/
//FILE:rgbycc.c                                                       */
//DATE:2006/8/4                                                       */
//example for FP-Fixer                                                */
/*Please type `Fpfix -all_*** -arg rgbdata.data rgbycc.c' on console. */
/**********************************************************************/

/*fixed module*/
void rgbycc(double r,double g,double b,
            double*y,double*cb, double*cr)
{
    *y = 0.299    * r + 0.587   * g + 0.114   * b;
    *cb= -0.16874 * r - 0.33126 * g + 0.5     * b + 0x80;
    *cr= 0.5      * r - 0.41869 * g - 0.08131 * b + 0x80;

}


/*for simulation code*/
void yccrgb(double y,double cb,double cr,
            double*r,double*g, double*b)
{
    *r = y + 1.402   * (cr - 0x80);
    *g = y - 0.34414 * (cb - 0x80) - 0.71414 * (cr - 0x80);
    *b = y + 1.772   * (cb - 0x80);
}


#include <stdio.h>
#include <stdlib.h>

double fpfix_error_rate=0.01;
double fpfix_error_rate2=0.2;

//#define _DEBUG_

double getData(FILE*fp);
int errJudge(int errflg,int totaldata,double acceptrate);
int checkErr(double r,double g,double b,
         double out_r,double out_g,double out_b, int type);

static int ng=0;

int main(int argc,char**argv)
{
    int i;
    double r,g,b;
    double y,cb,cr;
    double out_r,out_g,out_b;
    FILE*fp;

    if(argc!=2){
        printf("usage: <input-data>\n");
        return -1;
    }

    fp=fopen(argv[1],"r");
    if(!fp){
       perror(argv[1]);
       return -1;
    }

    i=0;
    ng=0;
    while((r=getData(fp))>=0.0){
      g=getData(fp);
      if(g<0.0) break;
      b=getData(fp);
      if(b<0.0) break;

      rgbycc(r,g,b,&y,&cb,&cr); //fpfix module
      yccrgb(y,cb,cr,&out_r,&out_g,&out_b);

#ifdef _DEBUG_
      printf ( "--- %5d ------------------------------------------\n", i );
#endif
      checkErr(r,g,b,out_r,out_g,out_b,1); //fpfix evFunc

      i++;
    }
    printf ( "total:%d ng:%d\n", i, ng );

    fclose ( fp );
    return 0;
}

double getData(FILE*fp)
{
    static char buf[1024];
    if(fgets(buf,1024,fp)==NULL) return -1.0;
    return atof(buf);
}

/* Evaluation function
 *
 *   Argument type
 *      0:% compares. : -max 40 used
 *      1: value compares.
 */

int checkErr_detail ( char *rgb, double exp, double out, int type )
{
   int flg=0;

    if( type == 0 ){
        // % compares. : -max 40 used
        double errate = fpfix_error_rate;
        double fMax=0.0;
        double fMin=0.0;
        double fExp=0.0;
        double fOut=0.0;

        fExp=exp;
        if ( fExp == 0.0 ) {
            fMax=(fExp + errate);
            fMin=(fExp - errate);
        } else if( fExp >= 0 ){
            fMax=(fExp * (1.0 + errate));
            fMin=(fExp * (1.0 - errate));
        } else {
            fMax=(fExp * (1.0 - errate));
            fMin=(fExp * (1.0 + errate));
        }

        fOut=out;

        flg = ( (fOut > fMax) || (fMin > fOut) );

#ifdef _DEBUG_
        sprintf ( szExp, "%lf", fExp );
        sprintf ( szMax, "%lf", fMax );
        sprintf ( szMin, "%lf", fMin );
        sprintf ( szOut, "%lf", fOut );

        printf ( " %s : %s : e.%10s o.%10s r.%lf | %10s %2s %10s %2s %10s\n",
            rgb,
            (flg != 0) ? "ng" : "  ",
            szExp, szOut, errate,
            szMax,
            ((fMax >= fOut) ? ">=" : "<"),
            szOut,
            ((fOut >= fMin) ? ">=" : "<"),
            szMin );
#endif
    } else {
        // value compares.
        double errate = fpfix_error_rate2;
        flg = ((exp+errate) < out) || ((exp-errate) > out );
    }

    if ( flg != 0 ) {
         ng++;
    }
    return ( flg );
}

int checkErr(double r,double g,double b,
         double out_r,double out_g,double out_b,
         int type)
{
   int flg=0;

   flg  = checkErr_detail ( "r", r, out_r, type );
   flg |= checkErr_detail ( "g", g, out_g, type );
   flg |= checkErr_detail ( "b", b, out_b, type );
   return ( flg );
}


